/*
	Genome-wide Efficient Mixed Model Association (GEMMA)
    Copyright (C) 2011  Xiang Zhou and Matthew Stephens

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __GEMMA_H__                
#define __GEMMA_H__

#include "gsl/gsl_vector.h"
#include "gsl/gsl_matrix.h"

#include "io.h"

using namespace std;



class FUNC_PARAM
{

public:
	bool calc_null;
	size_t ni_test;
	size_t n_cvt;
	const gsl_vector *eval;
	const gsl_matrix *Uab;
	const gsl_vector *ab;
	size_t e_mode;
};




class GEMMA {

public:
	//IO related parameters
	int a_mode;				//analysis mode, 1/2/3/4 for Frequentist tests
	int k_mode;				//kinship read mode: 1: n by n matrix, 2: id/id/k_value; 
	int e_mode;				//eigen mode: 0: exact; other number: approximation;
	
	int p_column;			//which phenotype column needs analysis
	int display_pace;		//display pace
	
	string file_bfile;
	string file_geno;
	string file_pheno;
	string file_anno;		//optional
	string file_cvt;		//optional
	string file_kin;
	string file_out;
	
	//SNP related parameters	
	double miss_level;
	double maf_level;	
	
	//Frequentist analysis parameters
	double l_min;
	double l_max;
	double n_region;
	double l_mle_null;
	double l_remle_null;

	// Other parameters
	bool error;
	size_t ni_total, ns_total;	//number of total individuals and snps
	size_t ni_test, ns_test;	//number of individuals and snps used for analysis
	size_t n_cvt;				//number of covariates
	double time_total;		//record total time
	double time_eigen;		//time spent on eigen-decomposition
	double time_Ux;			//time spent on calculating UX and Uy
	double time_opt;		//time spent on optimization iterations

	vector<double> pheno;					//a vector record all phenotypes, NA replaced with -9
	vector<vector<double> > cvt;			//a vector record all covariates, NA replaced with -9	
	vector<int> indicator_idv;				//indicator for individuals (phenotypes), 0 missing, 1 available for analysis
	vector<int> indicator_snp;				//sequence indicator for SNPs: 0 ignored because of (a) maf, (b) miss, (c) non-poly; 1 available for analysis
	vector<int> indicator_cvt;				//indicator for covariates, 0 missing, 1 available for analysis
	
	map<string, int> mapID2num;		//map small ID number to number, from 0 to n_phen-1
	map<string, int> mapRS2chr;		//map rs# to chromosome location
	map<string, long int> mapRS2bp;		//map rs# to base position
	
	vector<SNPINFO> snpInfo;		//record SNP information
	vector<SUMSTAT> sumStat;		//Output SNPSummary Data

	//Main functions
	GEMMA();
	void ReadFiles ();
	void ReadKin (gsl_matrix *G);
	void ReadKin (float *A);
	void AnalyzePlink (const gsl_matrix *U, const gsl_vector *eval, const gsl_matrix *UtW, const gsl_vector *Uty, const gsl_matrix *W, const gsl_vector *y);
	void AnalyzeBimbam (const gsl_matrix *U, const gsl_vector *eval, const gsl_matrix *UtW, const gsl_vector *Uty, const gsl_matrix *W, const gsl_vector *y);
	void WriteLog ();
	void WriteFiles ();
	
	void CalcKin (gsl_matrix *matrix_kin);
	void WriteKin (const gsl_matrix *matrix_kin);
	
	void CalcResidual (const double lambda, const gsl_matrix *U, const gsl_vector *eval, const gsl_matrix *UtW, const gsl_vector *Uty, gsl_vector *Pcy);
	void CalcUab (const gsl_matrix *UtW, const gsl_vector *Uty, gsl_matrix *Uab);
	void CalcUab (const gsl_matrix *UtW, const gsl_vector *Uty, const gsl_vector *Utx, gsl_matrix *Uab);
	void Calcab (const gsl_matrix *W, const gsl_vector *y, gsl_vector *ab);
	void Calcab (const gsl_matrix *W, const gsl_vector *y, const gsl_vector *x, gsl_vector *ab);
	void CalcLambda (const char func_name, FUNC_PARAM &params, double &lambda, double &logf);
	void CalcRLWald (const double &lambda, const FUNC_PARAM &params, double &beta, double &se, double &p_wald);
	void CalcRLScore (const double &l, const FUNC_PARAM &params, double &beta, double &se, double &p_score);
	
};


#endif


