/*
	Genome-wide Efficient Mixed Model Association (GEMMA)
    Copyright (C) 2011  Xiang Zhou and Matthew Stephens

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <string>
#include <cstring>
#include <sys/stat.h>
#include <cmath>


#include "gsl/gsl_vector.h"
#include "gsl/gsl_matrix.h"
#include "gsl/gsl_linalg.h"
#include "gsl/gsl_blas.h"
#include "gsl/gsl_eigen.h"

#ifdef WITH_ARPACK
#include "ardsmat.h"
#include "ardssym.h"
#endif	

#include "param.h"
#include "gemma.h"

#ifdef WITH_LAPACK
#include "lapack.h"
#endif

using namespace std;





void PARAM::PrintHeader (void)
{
	cout<<endl;
	cout<<"*********************************************************"<<endl;
	cout<<"* Genome-wide Efficient Mixed Model Association (GEMMA) *"<<endl;
	cout<<"* Version 0.92, April 2, 2012                           *"<<endl;
	cout<<"* Visit                                                 *"<<endl;
	cout<<"*    http://stephenslab.uchicago.edu/software.html      *"<<endl;
	cout<<"*    http://home.uchicago.edu/~xz7/software.html        *"<<endl;
	cout<<"* For Possible Updates                                  *"<<endl;
	cout<<"* (C) 2012 Xiang Zhou and Matthew Stephens              *"<<endl;
	cout<<"* GNU General Public License                            *"<<endl;
	cout<<"* For Help, Type ./gemma -h                             *"<<endl;
	cout<<"*********************************************************"<<endl;
	cout<<endl;
	
	return;
}


void PARAM::PrintLicense (void)
{
	cout<<endl;
	cout<<"The Software Is Distributed Under GNU General Public License, But May Also Require The Following Notifications."<<endl;
	cout<<endl;
	
	cout<<"Including Lapack Routines In The Software May Require The Following Notification:"<<endl;
	cout<<"Copyright (c) 1992-2010 The University of Tennessee and The University of Tennessee Research Foundation.  All rights reserved."<<endl;
	cout<<"Copyright (c) 2000-2010 The University of California Berkeley. All rights reserved."<<endl;
	cout<<"Copyright (c) 2006-2010 The University of Colorado Denver.  All rights reserved."<<endl;	
	cout<<endl;
	cout<<"Including Arpack Routines In The Software May Require The Following Notification:"<<endl;
	cout<<"Copyright (c) 1996-2008 Rice University."<<endl;
	cout<<"Developed by D.C. Sorensen, R.B. Lehoucq, C. Yang, and K. Maschhoff."<<endl;
	cout<<"All rights reserved."<<endl;	
	cout<<endl;
	
	cout<<"$COPYRIGHT$"<<endl;
	cout<<"Additional copyrights may follow"<<endl;
	cout<<"$HEADER$"<<endl;
	cout<<"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:"<<endl;
	cout<<"- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer."<<endl;
	cout<<"- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer listed in this license in the documentation and/or other materials provided with the distribution."<<endl;
	cout<<"- Neither the name of the copyright holders nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission."<<endl;
	cout<<"The copyright holders provide no reassurances that the source code provided does not infringe any patent, copyright, or any other "
		<<"intellectual property rights of third parties.  The copyright holders disclaim any liability to any recipient for claims brought against "
		<<"recipient by any third party for infringement of that parties intellectual property rights. "<<endl;
	cout<<"THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT "
		<<"LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT "
		<<"OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT "
		<<"LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY "
		<<"THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE "
		<<"OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."<<endl;
	cout<<endl;
	
	
	
	return;
}



void PARAM::PrintHelp(void)
{
	cout<<endl; 
	cout<<" GEMMA version 0.92, released on April 2, 2012 by X. Zhou and M. Stephens"<<endl; 
	cout<<" The Software Performs Association Tests Using A Standard Linear Mixed Model"<<endl; 
	cout<<endl;
	
	cout<<" QUICK GUIDE" << endl;
	cout<<" To generate kinship/relatedness matrix: "<<endl;
	cout<<"         ./gemma -bfile [prefix] -gk [num] -o [prefix]"<<endl;
	cout<<"         ./gemma -g [filename] -p [filename] -gk [num] -o [prefix]"<<endl;
	cout<<" To perform frequentist analysis: "<<endl;
	cout<<"         ./gemma -bfile [prefix] -k [filename] -fa [num] -o [prefix]"<<endl;
	cout<<"         ./gemma -g [filename] -p [filename] -a [filename] -k [filename] -fa [num] -o [prefix]"<<endl;	
	cout<<endl;
	
	cout<<" FILE I/O RELATED OPTIONS" << endl;
	cout<<" -bfile    [prefix]       "<<" specify input PLINK binary ped file prefix."<<endl;	
	cout<<"          requires: *.fam, *.bim and *.bed files"<<endl;	
	cout<<" -g        [filename]     "<<" specify input BIMBAM mean genotype file name"<<endl;
	cout<<"          format: rs#1, allele0, allele1, genotype for individual 1, genotype for individual 2, ..."<<endl;	
	cout<<"                  rs#2, allele0, allele1, genotype for individual 1, genotype for individual 2, ..."<<endl;	
	cout<<"                  ..."<<endl;	
	cout<<"          missing genotype value: NA"<<endl;	
	cout<<" -p        [filename]     "<<" specify input BIMBAM phenotype file name"<<endl;
	cout<<"          format: phenotype for individual 1"<<endl;	
	cout<<"                  phenotype for individual 2"<<endl;	
	cout<<"                  ..."<<endl;
	cout<<"          missing phenotype value: NA"<<endl;	
	cout<<" -a        [filename]     "<<" specify input BIMBAM SNP annotation file name (optional)"<<endl;	
	cout<<"          format: rs#1, base_position, chr_number"<<endl;	
	cout<<"                  rs#2, base_position, chr_number"<<endl;	
	cout<<"                  ..."<<endl;
	cout<<" -k        [filename]     "<<" specify input kinship/relatedness matrix file name"<<endl;	
	cout<<" -c        [filename]     "<<" specify input covariates file name (optional)"<<endl;	
	cout<<"          format: covariate 1 for individual 1, ... , covariate c for individual 1"<<endl;	
	cout<<"                  covariate 1 for individual 2, ... , covariate c for individual 2"<<endl;	
	cout<<"                  ..."<<endl;
	cout<<"          missing covariate value: NA"<<endl;	
	cout<<"          note: the intercept (a column of 1s) may need to be included"<<endl;
	cout<<" -km       [num]          "<<" specify input kinship/relatedness file type (default 1)."<<endl;
	cout<<"          options: 1: \"n by n matrix\" format"<<endl;
	cout<<"                   2: \"id  id  value\" format"<<endl;
	cout<<" -n        [num]          "<<" specify phenotype column in the phenotype/*.fam file (optional; default 1)"<<endl;	
	cout<<" -pace     [num]          "<<" specify terminal display update pace (default 100000 SNPs)."<<endl;
	cout<<" -o        [prefix]       "<<" specify output file prefix (default \"result\")"<<endl;  
	cout<<"          output: prefix.cXX.txt or prefix.sXX.txt from kinship/relatedness matrix estimation"<<endl;	
	cout<<"          output: prefix.assoc.txt and prefix.log.txt form association tests"<<endl;	
	cout<<endl;
	
	cout<<" SNP QC OPTIONS" << endl;
	cout<<" -miss     [num]          "<<" specify missingness threshold (default 0.05)" << endl; 
	cout<<" -maf      [num]          "<<" specify minor allele frequency threshold (default 0.01)" << endl; 
	cout<<endl;
	
	cout<<" RELATEDNESS MATRIX CALCULATION OPTIONS" << endl;
	cout<<" -gk       [num]          "<<" specify which type of kinship/relatedness matrix to generate (default 1)" << endl; 
	cout<<"          options: 1: centered XX^T/p"<<endl;
	cout<<"                   2: standardized XX^T/p"<<endl;
	cout<<"          note: non-polymorphic SNPs are excluded "<<endl;
	cout<<endl;
	
	cout<<" ASSOCIATION TESTS OPTIONS" << endl;		
	cout<<" -fa       [num]          "<<" specify frequentist analysis choice (default 1)."<<endl;
	cout<<"          options: 1: Wald test"<<endl;
	cout<<"                   2: Likelihood ratio test"<<endl;
	cout<<"                   3: Score test"<<endl;
	cout<<"                   4: 1-3"<<endl;
	cout<<" -lmin     [num]          "<<" specify minimal value for lambda (default 1e-5)" << endl; 
	cout<<" -lmax     [num]          "<<" specify maximum value for lambda (default 1e+5)" << endl; 
	cout<<" -region   [num]          "<<" specify the number of regions used to evaluate lambda (default 10)" << endl; 
	cout<<endl;
#ifdef WITH_ARPACK
	//	cout<<" -apprx    [num]          "<<" approximate the relatedness matrix using only the leading eigenvectors (default 1000)."<<endl;
	//	cout<<endl;
#endif
	cout<<" NOTE"<<endl;
	cout<<" 1. Only individuals with non-missing phenotoypes and covariates will be analyzed."<<endl;
	cout<<" 2. Missing genotoypes will be repalced with the mean genotype of that SNP."<<endl;
	cout<<" 3. Required memory size is roughly ten times the size of the kinship/relatedness matrix file."<<endl;
	//	cout<<" 4. Careful with -apprx option, as using a low-rank relatedness matrix may comprimise the model's ability to control for stratification."<<endl;
	cout<<endl;
	return;
}



void PARAM::Assign(int argc, char ** argv, GEMMA &cGemma)
{
	string str;
	
	for(int i = 1; i < argc; i++) {		
		if (strcmp(argv[i], "-bfile")==0 || strcmp(argv[i], "--bfile")==0 || strcmp(argv[i], "-b")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.file_bfile=str;
		}
		else if (strcmp(argv[i], "-g")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.file_geno=str;
		}
		else if (strcmp(argv[i], "-p")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.file_pheno=str;
		}
		else if (strcmp(argv[i], "-a")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.file_anno=str;
		}
		else if (strcmp(argv[i], "-k")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.file_kin=str;
		}
		else if (strcmp(argv[i], "-c")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.file_cvt=str;
		}
		else if (strcmp(argv[i], "-km")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.k_mode=atoi(str.c_str());
		}		
		else if (strcmp(argv[i], "-o")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.file_out=str;
		}		
		else if (strcmp(argv[i], "-fa")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {cGemma.a_mode=1; continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.a_mode=atoi(str.c_str());
		}		
		else if (strcmp(argv[i], "-lmin")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.l_min=atof(str.c_str());
		}
		else if (strcmp(argv[i], "-lmax")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.l_max=atof(str.c_str());
		}
		else if (strcmp(argv[i], "-region")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.n_region=atoi(str.c_str());
		}
		else if (strcmp(argv[i], "-n")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.p_column=atoi(str.c_str());
		}
#ifdef WITH_ARPACK
		else if (strcmp(argv[i], "-apprx")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {cGemma.e_mode=1000; continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.e_mode=atoi(str.c_str());
		}
#endif
		else if (strcmp(argv[i], "-gk")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {cGemma.a_mode=21; continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.a_mode=20+atoi(str.c_str());
		}
		else if (strcmp(argv[i], "-pace")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.display_pace=atoi(str.c_str());
		}
		else if (strcmp(argv[i], "-miss")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.miss_level=atof(str.c_str());
		}
		else if (strcmp(argv[i], "-maf")==0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') {continue;}
			++i;
			str.clear();
			str.assign(argv[i]);
			cGemma.maf_level=atof(str.c_str());
		}		
		else {cout<<"error! unrecognized option: "<<argv[i][0]<<argv[i][1]<<endl; cGemma.error=true; continue;}
	}
	
	return;
}




bool PARAM::CheckError (GEMMA &cGemma) {	
	struct stat fileInfo;
	string str;
	bool flag=false;		
	
	if (cGemma.k_mode!=1 && cGemma.k_mode!=2) {cout<<"error! unknown kinship/relatedness input mode: "<<cGemma.k_mode<<endl; flag=true;}
	if (cGemma.a_mode!=1 && cGemma.a_mode!=2 && cGemma.a_mode!=3 && cGemma.a_mode!=4 && cGemma.a_mode!=21 && cGemma.a_mode!=22)   
		{cout<<"error! unknown analysis mode: "<<cGemma.a_mode<<". make sure -fa or -gk is sepcified correctly."<<endl; flag=true;}
	if (cGemma.miss_level>1 || cGemma.miss_level<0) {cout<<"error! missing level needs to be between 0 and 1. current value = "<<cGemma.miss_level<<endl; flag=true;}
	if (cGemma.maf_level>1 || cGemma.maf_level<0) {cout<<"error! maf level needs to be between 0 and 0.5. current value = "<<cGemma.maf_level<<endl; flag=true;}
	if (cGemma.display_pace<0) {cout<<"error! display pace needs to be above 0. current value = "<<cGemma.display_pace<<endl; flag=true;}
	if (cGemma.l_max<cGemma.l_min) {cout<<"error! maximum lambda value must be larger than the minimal value. current values = "<<cGemma.l_max<<" and "<<cGemma.l_min<<endl; flag=true;}	
	
	if (!(cGemma.file_bfile).empty()) {
		if (!(cGemma.file_geno).empty()) {
			cout<<"error! only one genotype file is allowed. "<<endl; 
			cout<<"       current plink binary file prefix: "<<cGemma.file_bfile<<endl; 
			cout<<"       current bimbam mean genotype file name: "<<cGemma.file_geno<<endl; 
			flag=true;
		}
		else {
			str=cGemma.file_bfile+".bim";
			if(stat(str.c_str(),&fileInfo)==-1) {cout<<"error! fail to open .bim file: "<<str<<endl; flag=true;}
			str=cGemma.file_bfile+".bed";
			if(stat(str.c_str(),&fileInfo)==-1) {cout<<"error! fail to open .bed file: "<<str<<endl; flag=true;}
			str=cGemma.file_bfile+".fam";
			if(stat(str.c_str(),&fileInfo)==-1) {cout<<"error! fail to open .fam file: "<<str<<endl; flag=true;}
		}
	}
	else if ((cGemma.file_geno).empty()) {
		cout<<"error! missing genotype file name. "<<endl; 
		flag=true;
	}
	else if ((cGemma.file_pheno).empty()) {
		cout<<"error! missing phenotype file name. "<<endl; 
		flag=true;
	}	
	else {
		if (cGemma.k_mode==2) {cout<<"error! genotype file mode and kinship/relatedness file mode do not match. use \"-km 1\" when using bimbam mean genotype file. "<<endl; flag=true;}
	
		str=cGemma.file_geno;
		if(stat(str.c_str(),&fileInfo)==-1) {cout<<"error! fail to open bimbam mean genotype file: "<<str<<endl; flag=true;}
			
		str=cGemma.file_pheno;
		if(stat(str.c_str(),&fileInfo)==-1) {cout<<"error! fail to open bimbam phenotype file: "<<str<<endl; flag=true;}
		
		str=cGemma.file_anno;
		if (!str.empty() && stat(str.c_str(),&fileInfo)==-1) {
			cout<<"error! fail to open annotation file: "<<str<<endl; flag=true;
		}
	}
	
	str=cGemma.file_kin;
	if (cGemma.a_mode!=21 && cGemma.a_mode!=22) {
		if (str.empty() )  {cout<<"error! missing kinship/relatedness file name. "<<endl;  flag=true;}
		else if (stat(str.c_str(),&fileInfo)==-1) {cout<<"error! fail to open kinship/relatedness file: "<<str<<endl; flag=true;}
	}
	
	str=cGemma.file_cvt;
	if (!str.empty() && stat(str.c_str(), &fileInfo)==-1) {cout<<"error! fail to open covariate file: "<<str<<endl; flag=true;}	
	
	
	return flag;
}



void PARAM::BatchRun (GEMMA &cGemma) {
	clock_t time_begin, time_start;
	time_begin=clock();	
	
	cout<<"Reading Files ... "<<endl;
	cGemma.ReadFiles();
	if (cGemma.error==true) {cout<<"error! fail to read files. "<<endl; return;}
	
	//Generate Kinship matrix
	if (cGemma.a_mode==21 || cGemma.a_mode==22) {
		cGemma.ni_total=cGemma.indicator_idv.size();
		cGemma.ns_total=cGemma.indicator_snp.size();
		cout<<"## number of total individuals = "<<cGemma.ni_total<<endl;
		cout<<"## number of total SNPs = "<<cGemma.ns_total<<endl;
		cout<<"## number of analyzed SNPs = "<<cGemma.ns_test<<endl;
		cout<<"Calculating Relatedness Matrix ... "<<endl;
		
		gsl_matrix *matrix_kin=gsl_matrix_alloc (cGemma.ni_total, cGemma.ni_total);
		cGemma.CalcKin (matrix_kin);		
		if (cGemma.error==true) {cout<<"error! fail to calculate relatedness matrix. "<<endl; return;}
		
		cGemma.WriteKin (matrix_kin);		
		return;
	}
	
	//Perform Analysis
	if ((cGemma.file_cvt).empty() || (cGemma.indicator_cvt).size()==0) {
		cout<<"No covariate file is provided. Intecept will be used."<<endl; 
		cGemma.n_cvt=1;
	}
	if ( (cGemma.indicator_cvt).size()!=0 && (cGemma.indicator_cvt).size()!=(cGemma.indicator_idv).size()) {
		cout<<"error! number of rows in the covariates file do not match the number of individuals. "<<endl;
		return;
	}
	
	//calculate ni_total and ni_test, and set indicator_idv to 0 whenever indicator_cvt=0
	cGemma.ni_total=(cGemma.indicator_idv).size();
	cGemma.ni_test=0; 
	if ((cGemma.indicator_cvt).size()!=0) {
		for (vector<int>::size_type i=0; i<(cGemma.indicator_idv).size(); ++i) {
			if (cGemma.indicator_idv[i]!=0 && cGemma.indicator_cvt[i]==0) {cGemma.indicator_idv[i]=0;}
			if (cGemma.indicator_idv[i]==0) {continue;}
			cGemma.ni_test++;
		}
	}
	else {
		for (vector<int>::size_type i=0; i<(cGemma.indicator_idv).size(); ++i) {
			if (cGemma.indicator_idv[i]==0) {continue;}
			cGemma.ni_test++;
		}
	}	
	
	cGemma.ns_total=cGemma.indicator_snp.size();
	
	cout<<"## number of total individuals = "<<cGemma.ni_total<<endl;
	cout<<"## number of total SNPs = "<<cGemma.ns_total<<endl;
	cout<<"## number of analyzed individuals = "<<cGemma.ni_test<<endl;
	cout<<"## number of analyzed SNPs = "<<cGemma.ns_test<<endl;
	cout<<"## number of covariates = "<<cGemma.n_cvt<<endl;	
	
	gsl_matrix *W=gsl_matrix_alloc (cGemma.ni_test, cGemma.n_cvt);	
	gsl_vector *y=gsl_vector_alloc (cGemma.ni_test);
	
	//set covariates matrix W and phenotype vector y
	size_t ci_test=0;
	if (cGemma.indicator_cvt.size()!=0) {
		for (vector<int>::size_type i=0; i<cGemma.indicator_idv.size(); ++i) {
			if (cGemma.indicator_idv[i]==0 || cGemma.indicator_cvt[i]==0) {continue;}
			gsl_vector_set (y, ci_test, (cGemma.pheno)[i]);
			for (size_t j=0; j<cGemma.n_cvt; ++j) {
				gsl_matrix_set (W, ci_test, j, (cGemma.cvt)[i][j]);
			}
			ci_test++;
		}
	}
	else {
		gsl_matrix_set_all (W, 1.0);
		for (vector<int>::size_type i=0; i<cGemma.indicator_idv.size(); ++i) {
			if (cGemma.indicator_idv[i]==0) {continue;}
			gsl_vector_set (y, ci_test, (cGemma.pheno)[i]);	
			ci_test++;
		}
	}		
	
	//read relatedness matrix G and perform eigen decomposition
	gsl_matrix *U;
	gsl_vector *eval; 	
	
	time_start=clock();	
	if (cGemma.e_mode==0) {	
		gsl_matrix *G=gsl_matrix_alloc (cGemma.ni_test, cGemma.ni_test);
		U=gsl_matrix_alloc (cGemma.ni_test, cGemma.ni_test); 
		eval=gsl_vector_alloc (cGemma.ni_test);
		
		cGemma.ReadKin(G);
		if (cGemma.error==true) {cout<<"error! fail to read kinship/relatedness file. "<<endl; return;}
		
#ifdef WITH_LAPACK
		lapack_eigen_symmv (G, eval, U);
#else
		gsl_eigen_symmv_workspace *w=gsl_eigen_symmv_alloc (cGemma.ni_test);
		gsl_eigen_symmv (G, eval, U, w);
		gsl_eigen_symmv_free (w);	
#endif	

		for (size_t i=0; i<eval->size; ++i) {
			if (gsl_vector_get (eval, i)<1e-10) {
				gsl_vector_set (eval, i, 0);
			}
		}
		gsl_matrix_free (G);
		
	}
#ifdef WITH_ARPACK
	else {
		size_t n_sym=cGemma.ni_test*(cGemma.ni_test+1)/2;
		float *A=new float [n_sym];
		cGemma.ReadKin(A);
		
		ARdsSymMatrix<float> matrix(cGemma.ni_test, A);
		// Defining what we need: the eigenvectors of A with largest magnitude.
		ARluSymStdEig<float> dprob(cGemma.e_mode, matrix, "LM", 0, 1e-10, 1000);
		// Finding eigenvalues and eigenvectors.
		dprob.FindEigenvectors();
		
		delete [] A;
		
		U=gsl_matrix_alloc (cGemma.ni_test,dprob.ConvergedEigenvalues()); 
		eval=gsl_vector_alloc (dprob.ConvergedEigenvalues());
		gsl_matrix_set_zero (U); 
		gsl_vector_set_zero (eval);
		for (int i=0; i<dprob.ConvergedEigenvalues(); i++) {
			double d=dprob.Eigenvalue(i);
			if (fabs(d)>1e-10) {
				gsl_vector_set (eval, i, d);
				for (int j=0; j<dprob.GetN(); j++) { 
					gsl_matrix_set(U, j, i, dprob.Eigenvector(i,j));
				}
			}
			else {
				gsl_vector_set (eval, i, 0);
				gsl_vector_view U_col=gsl_matrix_column (U, i);
				gsl_vector_set_zero (&U_col.vector);
			}
		}		
	}
#endif	

	cGemma.time_eigen=(clock()-time_start)/(double(CLOCKS_PER_SEC)*60.0);
		
	
	//calculate UtW and Uty
	gsl_matrix *UtW=gsl_matrix_alloc (U->size2, cGemma.n_cvt);
	gsl_vector *Uty=gsl_vector_alloc (U->size2);
	
	for (size_t i=0; i<W->size2; ++i) {
		gsl_vector_view W_col=gsl_matrix_column (W, i);
		gsl_vector_view UtW_col=gsl_matrix_column (UtW, i);
		gsl_blas_dgemv (CblasTrans, 1.0, U, &W_col.vector, 0.0, &UtW_col.vector);
	}
	gsl_blas_dgemv (CblasTrans, 1.0, U, y, 0.0, Uty);
	
	//analysis
	if (!cGemma.file_bfile.empty()) {
		cGemma.AnalyzePlink (U, eval, UtW, Uty, W, y);
	}
	else {
		cGemma.AnalyzeBimbam (U, eval, UtW, Uty, W, y);
	}	
	
	cGemma.time_total=(clock()-time_begin)/(double(CLOCKS_PER_SEC)*60.0);
	
	cGemma.WriteLog();
	cGemma.WriteFiles();
	
	
	gsl_matrix_free (W);
	gsl_matrix_free (U);
	gsl_matrix_free (UtW);
	gsl_vector_free (eval);
	gsl_vector_free (y);
	gsl_vector_free (Uty);
	return;
}



